__code__='GatU0D/[on&>$QoP(Q(>!\\G+T/Mu[ib$@PAKFgJW7sL/dCNKfW*3Cd4ZUmi<[b9J0/lMHI;SVigorg*V[NPYX,^Qp]?LQ1`o?=:K67@<S2"/jLs,k)i0I06M18\'.V!3ZEp%0\\/r.p>XG!EB/r;$F$%ec?_?`rN";VZE9;`rSs9IKas/n.adhi7ELcJ>WG2!7(Y:"oJSo>JP<1k5i="]E!0gn9_rr694_5b#nE;^&WX_?s&%i*E7Y/W#DTq]WQj2q(Jg6YJG9HQcsJ@"FbkelUQiK$52Z5)./*[mi:4)$0>Wha=+tc!es_=%pH%H<_:eufk\'g]S.5Q)!m\\N[Jd"-,;.]2a^i[Cnb+5Q-FARPT:8p`U\'mT4%EOj:*HKI1YJ&1nqrC=QA/,*3UAWMu+;Dta[gH)J*#I.-Df+biup/mAUY$Z5tnYV4"YtTLT[O`Y>,V9MY$$2AbNn))5i7`3u\\\\5B&0fnY8>Fi$Y@ac35UbT<aI:o-_m*&/.\\S-k\'0dD!)-?&^/&r`(DgST<8V)XM2PneUZUAE=boTta)Qa^O.1k^eS\\&#k]dSoj>Q>5^h)fj2M16@W\\JW-@#,t=TP1slW39K*LHD-NLcl?)anp=e5;`kb,+9LirfHgZ0V:!]*\\+<CbG8e[V-W?tcDa^^]Mqa]Hs/=G(0;pTE9h,N][AkO>?[H\\ZUaQdj#X6)/h4pZO;Bu=1mRS\'[ff"NqBous8-!>Pk^!T`dSR3fm#YlWtS]i+0kP[,RpR>l@!*>m>Pbu]8c,X%J0Y:n>6EW+KrDCjZrNh2Wk,Kq5U<p:lcF.t17G0Lb*6mFmF&\'X+I,*RS?`A2Yjoumi2l*^qCU1C=KG^ekY35"etau`i^3M+j9mu%/$/[5u")+:aH_SI=kGt9+7.DEKncL4#P@ia6Nq(t]RQWVjE\\,bi+\\+lB0A:_r(KqR">*aZBH5_YJI"DS12g_!jN@GS<o[dLO?`dRSMNF<N)hs^01dr8;Nq;S\\tYgZ99M7q!M2+O.Z>*QP-l`%0\'q0imq6ibA1_h;9%MQ@^67q]lah7ho5s*q*+"b&OO-[44Jq8lY:l,L)t(6),fcW0*qU[f2M."57~>'



"""
This ReportLab sample demonstrates how to create dynamic documents
using RML and a mySQL database with the 'preppy' templating engine.
It creates a few simple one-page PDFs.

Other databases (odbc, Postgres etc) all follow a highly similar API - typically
the connect method is all that differs at this level.

Create the required database with the script 'create_rldb.sql', and look
inside 'rldb.prep' to see how the basic PDF document is described in Report
Markup Language.

If you're an old Python hand, probably the only portions relevant here
are the 'main()' function at the bottom and the 'rmltemplate.prep' file.
If new to the language, we'll digress for a little and discuss how the
language handles databases.

If you're sitting down to write a database report using Python, preppy and RML,
you may be pleasantly surprised how readable it can be.  Python has a standard
database API, so the connection code for MySQL, Postgresql, Oracle, ODBC and so
on tends to be almost identical.

When writing a report or generating a web page, we generally just want no-fuss
access to variables.  People from Java or old Microsoft days remember a lot of lines
of code with names like OpenRecordSet, FieldByName, MoveNext etc;
and a lot of brackets and quotes to type as you access fields by name or position.

In Python, when you do a query, you just grab some rows and get back a nested
structure like this    (the last item is a date object)

((1, 'John', 'Smith', datetime.date(1975, 1, 12)),
 (2, 'Jane', 'Bloggs', datetime.date(1977, 2, 1)),
 (3, 'Yamanda', 'Taro', datetime.date(1980, 3, 10)),
 (4, 'Nai', 'Gor', datetime.date(1985, 5, 20)))

If you're just looping over the data, this is often all you need.  You can access
things by indexing (thus, above, rows[0][2] is 'Smith'), but also easily loop
and unpack into named variables.  Thus, in an HTMl or RML template, this works:

{{for (id, firstName, surname, DOB) in rows}}
  <tr>
    <td>{{id}}</td>
    <td>{{firstName}}</td>
    ....
  </tr>
{{}}

Note that if you are coming from something like Crystal Reports, you are NOT
tied to one result set, or a fixed subquery structure.  You can grab any data
you like and assemble it into a any structure that helps, as you have a 
full-blown programming language at your disposal.  For example, you might want
to go off and do queries for related info on some rows but not others; if it's
a line or two, it can be done in the template easily.


If you have a more complex data set with several record types and columns, sometimes 
it's nice to create a lightweight 'objects' which you can access by attribute...
    print obj.firstName, obj.surname


The dynamic nature of the language means you can do this in a few lines of code
and don't need a class for every table in your database.  We used this pattern
below.

If doing more sophisticated database work regularly, you might end up using an
Object-Relational Mapper (ORM) designed to save you time on queries.  Thus,
if generating PDF inside a Django project, you'd be free to use Django's ORM
to fetch some objects and then pass these into our templates.   Or, you might
be using an independent ORM like SQL Alchemy.


When creating report templates (or web pages with data), the goal is usually
to keep the template as readable as possible.  Thus, accessing fields by name
is usually good.  And if you find yourself writing many lines of code to
(say) format someone's exam grade, you might be better making a Student
class with a method to compute that result, so that the template stays as
simple as

<td>{{student.displayGrade()}}</td> 


"""

import MySQLdb
import preppy
from rlextra.rml2pdf import rml2pdf
from rlextra.ers import fetchobj


def main():
    #connect to the database
    db = MySQLdb.connect(user="root", passwd="", db="rldb")
    #map the friends table into something we can use in python

    #fetch my friends, represented as objects with named attributes
    friends = fetchobj.queryToObjects(db,'select * from friends')
    
    #load the RML template into preppy
    template = preppy.getModule('rmltemplate.prep')
    
    #Now loop through the database results.
    for friend in friends:
        #pass the object into the preppy template to create an RML file
        rmlText = template.get(friend)
        
        #Now create the PDF by passing the RML file into RML2PDF
        pdfFileName = friend.firstName + '.pdf'
        rml2pdf.go(rmlText, outputFileName=pdfFileName)
        print('saved %s' % pdfFileName)
    
if __name__=='__main__':
    main()
    



if __name__=='__main__':
	for s in [
			'import sys',
			'from rlextra.rml2pdf import rml2pdf',
			'from reportlab import rl_config',
			'rml2pdf.main(quiet=rl_config.verbose==0,fn=sys.argv[1:])'
			]:
		exec(s)



from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.shapes import Drawing, Group, Line
from reportlab.graphics.widgetbase import TypedPropertyCollection
from reportlab.lib.colors import Color, _PCMYK_black
from reportlab.lib.attrmap import *
from reportlab.lib.units import cm
from reportlab.lib.validators import *
from reportlab.pdfbase.pdfmetrics import getFont
from reportlab.graphics.charts import barcharts
from reportlab.graphics.charts.lineplots import _maxWidth

def _highest(data, roundInterval=10):
    v=max(list(map(max,data)))
    d,f=divmod(v,roundInterval)
    if f>0:
        f=1
    else:
        f=0.5
    return (d+f)*roundInterval

class HorizontalBarChart(barcharts.HorizontalBarChart):
    """Returns a widget - used by the class HorizontalBarChartDrawing (in file horizontalbarchart_db.py).
    Based on (and inherits from) barcharts.HorizontalBarChart.

    labelRow2 is an optional attribute - this can safely be set to
    None, in which case it will be ommited. If it does exist, it is
    used to produce a third column of labels for eg, 'previous quarter
    in brackets'."""

    _attrMap = AttrMap(BASE=barcharts.HorizontalBarChart,
        categoryNames = AttrMapValue(isNoneOrListOfNoneOrStrings,
            desc='List of category names'),
        background = AttrMapValue(isColorOrNone,
            desc='Background colour for chart area (if any)'),
        labelRow2 = AttrMapValue(isNoneOrListOfNoneOrNumbers,
            desc='List of numbers used for optional second set of labels (eg previous quarter)'),
        lineLength = AttrMapValue(isNumber,
            desc='Length of the horizontal line (behind the actual bars)'),
        lineStrokeWidth = AttrMapValue(None,
            desc='Stroke width of the horizontal line (behind the actual bars)'),
        lineColor = AttrMapValue(isColorOrNone,
            desc='Colour of the horizontal line (behind the actual bars)'),
        labelData = AttrMapValue(isAnything,
            desc="Data used for the labels"),
        labelFontName = AttrMapValue(isString,
            desc="Name of font used for the labels"),
        labelFontSize = AttrMapValue(isNumber,
            desc="Size of font used for the labels"),
        labelOffset = AttrMapValue(isNumber,
            desc="Space between label text and grid edge"),
        labelFillColor = AttrMapValue(isColor,
            desc="Ink color for the labels"),
        labelFormat = AttrMapValue(isAnything,
            desc="Formatting routine to be used on 2nd and 3rd columns of labels"),
        labelRightPadding = AttrMapValue(isAnything,
            desc="Space between 2nd and 3rd columns of labels"),
        leftPadding = AttrMapValue(isNumber,
            desc='Padding on left of drawing'),
        rightPadding = AttrMapValue(isNumber,
            desc='Padding on right of drawing'),
        topPadding = AttrMapValue(isNumber,
            desc='Padding at top of drawing'),
        bottomPadding = AttrMapValue(isNumber,
            desc='Padding at bottom of drawing'),
        valueMin = AttrMapValue(isNumber,
            desc='The lowest value to appear on the chart'),
        topOffset = AttrMapValue(isNumber, desc='the top of the enclosing drawing'),
        _labels = AttrMapValue(isListOfStrings,
            desc=''),
        _dxl = AttrMapValue(isNumber,
            desc=''),
        )

    def __init__(self):
        barcharts.HorizontalBarChart.__init__(self)
        self.barSpacing = 0.1*cm
        self.groupSpacing = 0.15*cm
        self.topOffset = 105

        self.labelFontName = "Helvetica-Oblique"
        self.labelFontSize = 6
        self.labelFillColor = _PCMYK_black
        self.labelOffset = 5
        self.labelRightPadding = 0.4*cm

        self.x = 0
        self.y = 0
        self.width = 1.5*cm # width of the BARS
        self.height = 3.5*cm # height of the BARS

        self.leftPadding=5
        self.topPadding=5

        # this defines a series of points - used for drawing bars and label1
        # self.data(0) is the data used to plot the bars
        # NB These go from bottom to top...
        self.data      = [(3.5, 15.5, 6.0, 7.5, 7.7,  7.7,  9.4, 10.9, 14.9, 16.9)]
        self.labelData = [(3.5, 15.5, 6.0, 7.5, 7.7,  7.7,  9.4, 10.9, 14.9, 16.9)]
        self.labelRow2 = [5.3,  3.3, 2.4, 4.5, 6.3, 10.1, 14.2, 12.5, 15.4, 26.2]
        self.categoryNames = ('Cash',
                              'Other',
                              'Italy',
                              'Switzerland',
                              'Sweden',
                              'Finland',
                              'Netherlands',
                              'Ireland',
                              'France',
                              'Germany')
        self.labelFormat = '%.1f%%'
        self.valueMin = 0

        # used for the horizontal line behind the bars
        self.lineLength = 1.5*cm
        self.lineStrokeWidth = 0.5
        self.lineColor = _PCMYK_black

        # control bar spacing. if useAbsolute = 1 then
        # the next parameters are in points; otherwise
        # they are 'proportions' and are normalized to
        # fit the available space.  Half a barSpacing
        # is allocated at the beginning and end of the
        # chart.
        self.useAbsolute = 1
        self.barWidth = 0.175*cm
        self.barSpacing = 0.185*cm
        self.valueAxis.visible = 0
        self.categoryAxis.visible = 0

        self.barLabelFormat = None

        self.bars = TypedPropertyCollection(barcharts.BarChartProperties)
        self.bars.strokeColor = _PCMYK_black
        self.background=colors.Color(254/255.0, 251/255.0, 242/255.0)
        self.background=None

        for f in range (0,len(self.data)):
            self.bars[f].fillColor = colors.Color(232/255.0,224/255.0,119/255.0)
            self.bars[f].strokeWidth = 0.5

        self._labels=[]
        if self.labelData != None and max(self.labelData)!=None:
            for f in range(0, len(self.labelData[0])):
                if self.labelFormat is None:
                    self._labels.append(self.labelData[0][f])
                elif isinstance(self.labelFormat,str):
                    self._labels.append(self.labelFormat % self.labelData[0][f])
                elif iscallable(self.labelFormat):
                    self._labels.append(self.labelFormat(self.labelData[0][f]))
        else:
            templabelData = []
            for f in range(0, len(self.data[0])):
                templabelData.append(None)
            self.labelData.append(tuple(templabelData))

        for f in range(0, len(self.data[0])):
            ty = self.y+(self.barWidth/2.0)+((self.barWidth+self.barSpacing)*f)
            self._dxl = _maxWidth(self.categoryNames, self.labelFontName, self.labelFontSize)           
            

    def draw(self): 
        g = Group()
        if self.categoryNames == None:
            self.categoryNames =[]
            for f in range(0, 10):
                self.categoryNames.append(None)
            self.categoryNames=tuple(self.categoryNames)
        if self.labelData != None and max(self.labelData[0])!=None:
            self.labelData = self.data

        # is labelRow2 empty?
        allNull = 0
        if self.labelRow2:
            for f in range (0,len(self.labelRow2)):
                if self.labelRow2[f]==None:
                    allNull=allNull+1
            if allNull==len(self.labelRow2):
                self.labelRow2 = None

        # the actual bar chart
        self.valueAxis.valueMin = self.valueMin
        # so we can actually see some of the line even for the longest bar
        self.valueAxis.valueMax = _highest(self.data)
        self.height = self._desiredCategoryAxisLength()
        self.y = self.topOffset - self.topPadding - self.height
        _dxl = _maxWidth(self._labels, self.labelFontName, self.labelFontSize) # this is NOT self.dxl!
        g.add(barcharts.HorizontalBarChart.draw(self))

        # do the labels
        ascent=getFont(self.labelFontName).face.ascent
        if ascent==0:
            ascent=0.718 # default (from helvetica)
        ascent=ascent*self.labelFontSize # normalize

        self._labels=[]
        if self.labelData != None and max(self.labelData[0])!=None:
            for f in range(0, len(self.labelData[0])):
                if self.labelFormat is None:
                    self._labels.append(self.labelData[0][f])
                elif isinstance(self.labelFormat,str):
                    self._labels.append(self.labelFormat % self.labelData[0][f])
                elif iscallable(self.labelFormat):
                    self._labels.append(self.labelFormat(self.labelData[0][f]))
        else:
            self.labelData=[]
            templabelData = []
            for f in range(0, 10):
                templabelData.append(None)
                self._labels.append(None)
            self.labelData.append(tuple(templabelData))

        R=list(range(0, len(self.data[0])))
        if self.categoryAxis.reverseDirection:
            R.reverse()

        ty = self.y+(self.barWidth+self.groupSpacing)/2.0
        tyInc = self.barWidth+self.groupSpacing
        for f in R:
            #left labels - column 1
            l = Label()
            l.setOrigin(self.x-self.labelOffset, ty)
            l.setText(self.categoryNames[f])
            l.fontName = self.labelFontName
            l.fontSize = self.labelFontSize
            l.fillColor = self.labelFillColor
            l.textAnchor= 'end'
            l.boxAnchor="e"
            g.add(l)

            #right labels - column 2
            l = Label()
            l.setOrigin(self.x+self.lineLength+self.labelOffset, ty)
            l.setText(self._labels[f])
            l.fontName = self.labelFontName
            l.fontSize = self.labelFontSize
            l.fillColor = self.labelFillColor
            l.boxAnchor="w"
            l.textAnchor= 'start'
            g.add(l)

            #right labels - column 3
            if self.labelRow2:
                l = Label()
                l.setOrigin(self.x+self.lineLength+self.labelOffset+_dxl+self.labelRightPadding, ty)
                if self.labelFormat is None:
                    labelText = self.labelRow2[f]
                elif isinstance(self.labelFormat,str):
                    labelText = self.labelFormat % self.labelRow2[f]
                elif iscallable(self.labelFormat):
                    labelText = self.labelFormat(self.labelRow2[f])
                else:
                    raise ValueError("Unknown formatter type %s, expected string or function" % labelFormat)
                labelText = "(%s)" % str(labelText)
                l.setText(labelText)
                l.fontName = self.labelFontName
                l.fontSize = self.labelFontSize
                l.fillColor = self.labelFillColor
                l.boxAnchor="w"
                l.textAnchor= 'start'
                g.add(l)

            # the background lines - now done as lines
            le=self.data[0][f]*float(self.lineLength/_highest(self.data))
            g.add(Line(self.x+le, ty, self.x+self.lineLength, ty,
                       strokeColor = self.lineColor,
                       strokeWidth = self.lineStrokeWidth))
            ty=ty+tyInc

        g.shift(self.leftPadding+self._dxl+self.labelOffset, 0)
        return g

    def demo(self, drawing=None):
        if not drawing:
            drawing = Drawing(140,105)
        drawing.add(self.draw())
        return drawing            


if __name__ == "__main__":
    d = HorizontalBarChart()
    d.demo().save(fnRoot="d_horizontalbarchart")



"""map xml using into other formats.

A MapController maps one XML format to another by recursive tag
translation.

BASIC USAGE

In basic usage initialize a Map Controller

    M = MapController()

and assign substitution strings or MapNodes for each tag.

Then process text using

    translatedtext = M.process(originaltext)

There will be errors if the xml in originaltext do not
match the assigned substitution strings or MapNodes for M.

TOP LEVEL SUBSTITUTION:

Every MapController must have a top level substitution which
defines a substitution location for "%(__content__)s", for
example as in

    M[''] = ''' <html>
         %(__content__)s
         </BODY>
         </html>
         '''

Here the empty tag '' signifies the top level.  In this particular
example the start of the <BODY> tag might be introduced in the translation
of another tag (like <DOCUMENTTITLE>).

SUBSTITUTION STRINGS:

You can either assign a substitution string to a tag name (for
tags with content) or assign a MapNode for tags which either do
not have content or require other special handling (such as
attribute defaults and transforms).

FOR TAGS WITH CONTENT, you can directly assign a string (which
internally is converted into a MapNode), for example

    M["H1"] = "<br><br><font size=+3><b>%(__content__)s</b></font><br>"

In this case the H1 tag has content (<H1>this is the content</H1>) but
no attributes that effect the translation.  Note that the marker
"%(__content__)s" indicates the location of the recursively translated
content.
    <H1>This is the content</H1>
translates to
    <br><br><font size=+3><b>This is the content</b></font><br>

FOR TAGS WITHOUT CONTENT or which require special handling assign
an explicit MapNode

    M["AUTHOR"] = MapNode(None, "<em>Aaron Watters</em>")

the AUTHOR tag has no content and no attributes that effect translation.
    <AUTHOR/>
translates to
    <em>Aaron Watters</em>

CONTAINED TAGS ARE RECURSIVELY TRANSLATED.  Using the translations
declared above
    <H1><AUTHOR/></H1>
becomes
    <br><br><font size=+3><b><em>Aaron Watters</em></b></font><br>

FOR TAGS WITH ATTRIBUTES include the appropriate attributes in the substition
string, as required.  For example,

    M["RETIRED"] = "<em>%(TITLE)s</em> <b>%(__content__)s</b> <em>Emeritas</em>"

The RETIRED tag has content and one attribute (TITLE) which is
required and effects translation.
    <RETIRED TITLE="Dr.">Carl Fungus</RETIRED>
becomes
    <em>Dr.</em> <b>Carl Fungus</b> <em>Emeritas</em>

For

    M["EXCLAMATION"] = "<FONT COLOR="RED"><B>%(TEXT)</B>, <EM>%(RECIPIENT)s</EM></FONT>"

EXCLAMATION has no content and two attributes (TEXT and RECIPIENT) which are required
and effect formatting.
    <EXCLAMATION TEXT="Holy rusty metal" RECIPIENT="Batman"/>
becomes
    <FONT COLOR="RED"><B>Holy rusty metal</B>, <EM>Batman</EM></FONT>

ATTRIBUTE DEFAULTS can be assigned using a MapNode assigned to a tag.
For example an alternate RETIRED tag might assign "Dr." as the default
value of Title.

    R = MapNode("<em>%(TITLE)s</em> <b>%(__content__)s</b> <em>Emeritas</em>")
    R.default["TITLE"] = "Dr."
    M["RETIRED"] =  R

Then using AUTHOR and RETIRED
    <RETIRED><AUTHOR/></RETIRED>
becomes
    <em>Dr.</em> <b><em>Aaron Watters</em></b> <em>Emeritas</em>
(Note the default "Dr." substitution).  I wish ;(.

ATTRIBUTE TRANSFORMS can be assigned using a MapNode.  For
example to upcase the titles of all retirees:

    R = MapNode("<em>%(TITLE)s</em> <b>%(__content__)s</b> <em>Emeritas</em>")
    R.addTransform(TITLE, str.upper)
    M["RETIRED"] =  R

Then
    <RETIRED TITLE="Dr.">Carl Fungus</RETIRED>
becomes
    <em>DR.</em> <b>Carl Fungus</b> <em>Emeritas</em>
The MapNode.transformContent(transform) will perform a transformation
on the recursively generated content of a tag (be careful when using
this one!)

MapNodes can alter the translation definitions for tags encountered
in their content.  For example

    M["Title"] = "<H1>%(__content__)s</H1>"
    ...
    C = MapNode("<hr>Chapter<hr> %(__content__)s")
    C["Title"] = "<H3>%(__content__)s</H3>"
    M["Chapter"] = C

Then a title outside of the Chapter context defaults to H1 but
a title inside a Chapter context defaults to H3.

"TAG NAME" OF None provides a "default" substitution
to apply to unknown tags, as in

    M[None] = "" # erase unknown tags
    M[None] = "#(__content__)s" # echo content for unknown tags

A substitution for a parent can refer to a substitution for a child

    # ref to %(TITLE.__content__)s refers to the title of the play (not act or scene)
    M["PLAY"] = ''' <html><head><title>%(TITLE.__content__)s</title></head><BODY>
                <CENTER><H1><FONT FACE="Arial">%(PLAY.TITLE.__content__)s</FONT></H1></CENTER>
         %(__content__)s
         <BODY>
         </html>
         '''

Any substitution can also refer to a "dotted path" to a content in the tree,
but it will get the last one if there is more than one match:

    M["ACT"]["TITLE"] = "<CENTER>%(PLAY.TITLE.__content__)s <H2>%(__content__)s</H2></CENTER>"

Here the top level element is "PLAY".  All "paths" must start at the top level element.

When "non local" uses of an elements __content__ are in use it is likely that the substitution
for the element itself will "vanish"

    M["TITLE"] = "" # __content__ used in top level (element vanishes)

__attrs__ entry allows for special handling of attributes
from rlextra.radxml.xmlmap import *
M=MapController()
M['']='%(__content__)s'
M['b']=MapNode('<B%(__attrs__)s>%(__content__)s</B>')
M['a']='<A>%(__content__)s</A>'

def bAttrsFunc(name,value):
    if name=='ZZ':
        return '--HACKED %s: %s' % (name,value)
    elif name=='x':
        return '--REVERSED %s: %s' %(name,''.join(reversed([x for x in value])))
    return value

M['b'].addTransform('__attrs__',AttrsFilter(rename=dict(z='ZZ'),remove=['y'],valueFunc=bAttrsFunc))
print M.process('<a>a contents 0<b z="30" y=\'"quoted"\' x="forward" w="leave alone">b content</b>a contents 1</a>')

"""
#" for emacs
# SET TO RUN HAMLET IN TEST SUITE
from __future__ import unicode_literals
DOHAMLET = 1
from reportlab.lib.utils import strTypes, asUnicode, asUnicodeEx, isPy3

class MapController:
    """Top level controller"""
    joinSeparator = ""

    def __init__(self, topLevelSubstitution = "%s", naive=0, eoCB=None, parseOpts={}):
        self.naive = naive # use nonvalidating parser if naive set
        self.topLevelSubstitution = topLevelSubstitution
        self.topDict = {}
        self.nodemappers = {None: None} # default
        self.eoCB = eoCB
        self._parseOpts = parseOpts
        self.preserve = PreserveNode()

    def __setitem__(self, item, value):
        if isinstance(value,strTypes):
            # very simple case, no defaults, no empty case
            value = MapNode(asUnicode(value))
        if isinstance(item,strTypes):
            item = asUnicode(item)
        self.nodemappers[item] = value

    def __getitem__(self, item):
        return self.nodemappers[item]

    def process(self, input, isTupleTree=False):
        if not isTupleTree:
            if self.naive:
                from reportlab.lib.rparsexml import parsexmlSimple as parsexml
            else:
                from reportlab.lib.rparsexml import parsexml
            elts = parsexml(input,eoCB=self.eoCB,parseOpts=self._parseOpts)
        else:
            elts = input
        return self.processParsed(elts)

    def processParsed(self, elts): # added to allow messing with the parsed input if needed
        D = self.topDict = {}
        D["__prefix__"] = ""
        elts = self.clean_elts(elts)
        textout = self.processContent(elts, top_level=1)
        T = asUnicodeEx(textout)
        result = self.topLevelSubstitution % T
        # clean out the topDict
        for k in list(D.keys()):
            del D[k] # possible self reference
        self.topDict = {}
        return result

    def processContent(self, elts, overrides=None, containerdict=None, top_level=1):
        if containerdict is None:
            containerdict = self.topDict
        if overrides is None:
            overrides = {} # no overrides yet
        if isinstance(elts,strTypes):
            return self.processString(asUnicode(elts))
        if isinstance(elts,tuple):
            return self.processTuple(elts, overrides, containerdict)
        else:
            L = []
            for e in elts:
                if isinstance(e,strTypes):
                    e2 = self.processString(asUnicode(e))
                else:
                    if not isinstance(e,tuple):
                        raise ValueError("bad content type %s" % type(e))
                    e2 = self.processTuple(e, overrides, containerdict)
                L.append(e2)
            return self.joinTranslatedContent(L)

    def processTuple(self, e, overrides, containerdict=None):
        tagname = e[0]
        nodemappers = self.nodemappers
        if overrides:
            nodemappers = nodemappers.copy()
            # do overrides
            nodemappers.update(overrides)
        defaultmapper = nodemappers.get(None, None)
        processor = nodemappers.get(tagname, defaultmapper)
        if processor is None:
            raise NameError("no processor for %s" % ascii(tagname))
        e2 = processor.translate(e, self, overrides, containerdict)
        return e2

    def joinTranslatedContent(self, L):
        return DelayedJoin(L, self.joinSeparator)

    def processString(self, elts):
        return elts # default, for now

    def clean_elts(self, elts):
        """optionally do stuff like dispose of all white content"""
        return elts

class DelayedJoin:
    value = None
    def __init__(self, L, s):
        self.L = L
        self.s = s

    if isPy3:
        def __str__(self):
            if self.value is not None: return self.value
            L = [asUnicodeEx(l) for l in self.L]
            self.value = result = self.s.join(L)
            # eliminate possible reference loops
            self.L = None
            return result

        def __bytes__(self):
            return self.__unicode__().encode('utf8')
    else:
        def __unicode__(self):
            if self.value is not None: return self.value
            L = [asUnicodeEx(l) for l in self.L]
            self.value = result = self.s.join(L)
            # eliminate possible reference loops
            self.L = None
            return result

        def __str__(self):
            return self.__unicode__().decode('utf8')

class ExtendedTransform:
    def __call__(self, t, sdict):
        return t

class SpecialTransform:
    '''transform that is given the name of the sdict entry so it can access the value directly'''
    def __call__(self, sdict, name):
        return asUnicodeEx(sdict[name])

class AttrsFilter(SpecialTransform):
    def __init__(self,rename={},remove=[],valueFunc=None):
        self.rename = rename
        self.remove = remove
        self.valueFunc = valueFunc
    def __call__(self, sdict, name):
        D = sdict[name]._D.copy()
        for k in self.remove:
            try:
                del D[k]
            except KeyError:
                pass
        for k,v in self.rename.items():
            try:
                D[v] = D.pop(k)
            except KeyError:
                pass
        valueFunc = self.valueFunc
        if valueFunc:
            for k,v in D.items():
                D[k] = valueFunc(k,v)
        return LazyAttrs(D)

class LazyKW:
    jstr = ', '
    pfx = ''
    def __init__(self,D):
        self._D = D
    if isPy3:
        def __str__(self):
            R = []
            for k,v in list(self._D.items()):
                R.append('%s="%s"' % (k,v))
            return self.pfx+self.jstr.join(R)

        def __bytes__(self):
            return self.__str__().encode('utf8')
    else:
        def __unicode__(self):
            R = []
            for k,v in list(self._D.items()):
                R.append('%s="%s"' % (k,v))
            return self.pfx+self.jstr.join(R)

        def __str__(self):
            return self.__unicode__().encode('utf8')

class LazyAttrs(LazyKW):
    jstr = ' '
    pfx = ' '

class MapNode(MapController):

    def __init__(self, full_substitution_string, empty_substitution_string=None, defaults=None):
        self.full_substitution_string = full_substitution_string
        self.empty_substitution_string = empty_substitution_string
        self.nodemappers = {}
        if defaults is None:
            defaults = {}
        self.defaults = defaults
        # transforms to apply per attribute to the content (eg make it have 2 digits and put a dollar in front)...
        self.transforms = {}

    def addTransform(self, name, t):
        self.transforms[name] = t

    def transformContent(self, t):
        self.addTransform('__content__',t)

    def translate(self, nodetuple, controller, overrides, containerdict=None):
        # add own overrides if present
        if self.nodemappers:
            overrides = overrides.copy()
            overrides.update(self.nodemappers)
        (tagname, attdict, content, extra) = nodetuple
        #tagname = nodedict[0]
        #content = nodedict.get(1, None)
        if not attdict: attdict = {}
        sdict = attdict.copy() # for modification
        prefix = tagname
        # prefix must be set before processing children
        #print "processing", tagname
        #print containerdict
        if containerdict is not None:
            cprefix = containerdict["__prefix__"]
            if cprefix:
                prefix = "%s.%s" % (cprefix, prefix)
        else:
            cprefix = ''
        sdict['__prefix__'] = prefix
        sdict['__parent_prefix__'] = cprefix
        sdict['__depth__'] = prefix.count('.')
        sdict['__parent_depth__'] = sdict['__depth__']-1
        sdict['__tag_depth__'] = (prefix+'.').count(tagname+'.')
        sdict['__mapnode__'] = self
        sdict['__controller__'] = controller
        sdict['__overrides__'] = overrides
        sdict['__containerdict__'] = containerdict
        sdict['__nodetuple__'] = nodetuple
        sdict["__tagname__"] = tagname
        sdict["__attrs__"] = LazyAttrs(attdict)
        defaults = self.defaults
        for name in list(defaults.keys()):
            if name not in sdict:
                sdict[name] = defaults[name]
        shadow = ShadowDict(tagname, sdict, containerdict)
        sdict['__shadow__'] = shadow
        #if content is None: stop
        if content is not None:
            pcontent = self.MyProcessContent(content, controller, overrides, shadow)
            sdict["__content__"] = pcontent
            # you can refer to TITLE.__content__
            top = controller.topDict
            if containerdict is not None:
                containerdict[tagname+".__content__"] = pcontent
            # you can refer to PLAY.TITLE.__content__
            top[prefix+".__content__"] = pcontent
            sstring = self.full_substitution_string
            if sstring is None:
                raise ValueError("no content allowed for %s" % ascii(tagname))
        else:
            sstring = self.empty_substitution_string
            if sstring is None:
                raise ValueError("content required for %s" % ascii(tagname))
        transforms = self.transforms
        for name in list(transforms.keys()):
            if name in sdict:
                t = transforms[name]
                if isinstance(t,SpecialTransform) or name=='__attrs__':
                    sdict[name] = t(sdict,name)
                else:
                    oldvalue = asUnicodeEx(sdict[name]) # IS STRING CONVERSION ALWAYS RIGHT?
                    if isinstance(t,ExtendedTransform):
                        sdict[name] = t(oldvalue, sdict)
                    else:
                        sdict[name] = t(oldvalue)
        try:
            result = shadow.substitute(sstring) #sstring % sdict
        except:
            raise ValueError("for tagname %s bad string %s for dict %s" % (
                ascii(tagname), ascii(sstring), ascii(list(sdict.keys()))))
        return result

    def MyProcessContent(self, content, controller, overrides, containerdict):
        """by default ask the global controller to do it"""
        return controller.processContent(content, overrides, containerdict)

class PreserveNode(MapNode):
    """This is a MapNode which aims to write out something
    close to the original tag"""

    def __init__(self):
         MapNode.__init__(self, None, None, None)

    def translate(self, nodetuple, controller, overrides, containerdict=None):

        #we want to leave this node unchanged, but let children through
        (tagName, attrs, content, extra) = nodetuple

        chunks = ['<%s' % tagName]
        if attrs is not None:
            for key, value in list(attrs.items()):
                chunks.append('%s="%s"' % (key, value))
        if content is None:
            chunks.append('/>')
        else:
            #must do children
            chunks.append('>')
            processedContent = self.MyProcessContent(content, controller, overrides, containerdict)
            chunks.append(asUnicodeEx(processedContent))
            chunks.append('</%s>' % tagName)
        return ''.join(chunks)

class ShadowDict:
    def __init__(self, name, dict=None, parent=None):
        self.name = name
        self.parent = {}
        self.dict = {}
        if parent is not None:
            self.parent = parent
        if dict is not None:
            self.dict = dict
    def __repr__(self):
        """tag like repr (but don't include parent info)"""
        n = self.name
        c = self.get("__content__", None)
        L = ["<%s" % n]
        d = self.dict
        K = list(d.keys())
        for name in K:
            if name[:2]!="__":
                v = d[name]
                L.append(' %s="%s"' % (name, v))
        if c is None:
            L.append("/>")
        else:
            L.append(">")
            L.append(c)
            L.append("</%s>" % n)
        return ''.join(L)
    def __getitem__(self, item):
        try:
            return self.dict[item]
        except KeyError:
            # special case: __repr__
            if item=="__repr__":
                return repr(self)
            return self.parent[item]
    def __setitem__(self, item, value):
        self.dict[item] = value
    def __contains__(self,item):
        return item in self.dict
    def get(self, item, default):
        try:
            return self[item]
        except:
            return default
    def namedsetitem(self, name, item, value):
        if self.name == name:
            self[item] = value
        else:
            try:
                self.parent.namedsetitem(name, item, value)
            except:
                raise ValueError("can't match name %s" % ascii(name))
    def substitute(self, s):
        return LazySubstitution(s, self)

class LazySubstitution:
    value = None
    def __init__(self, fmt, dictionary):
        self.fmt = fmt
        self.dictionary = dictionary

    if isPy3:
        def __str__(self):
            if self.value is not None: return self.value
            self.value = result = self.fmt % self.dictionary
            # now eliminate possible reference loops
            self.dictionary = None
            return result

        def __bytes__(self):
            return self.__str__().encode('utf8')
    else:
        def __str__(self):
            return self.__unicode__().encode('utf8')

        def __unicode__(self):
            if self.value is not None: return self.value
            self.value = result = self.fmt % self.dictionary
            # now eliminate possible reference loops
            self.dictionary = None
            return result

# hex colors for html
LIMEGREEN = "#32CD32"
LEMONCHIFFON = "#FFFACD"
LIGHTCORAL = "#F08080"
SKYBLUE = "#87CEEB"
CORNSILK = "#FFF8DC"

def textsection(title):
    L = [
    '<TABLE CELLSPACING="8" CELLPADDING="3" BORDER="0" WIDTH="98%%">',
    '<tr><TH bgcolor=%s align="right"><font face="arial">%s</font></td></tr>' % (LIMEGREEN, title),
    '<TR><TD><font face="arial">%(__content__)s</font></TD></TR>',
    '</TABLE>']
    return ''.join(L)

def customer_entry(title):
    return """<tr><td align="right" bgcolor="%s">
              <font face="arial"><em>%s</em></font></td>
        <td bgcolor="%s">
              <font face="arial"><b>%s</b></font></td></tr>""" % (
                  LEMONCHIFFON, title, LIGHTCORAL, "%(__content__)s")

charges_table_format = """
<center><TABLE BORDER="0" width=70%%>""" + """
<tr><td bgcolor=#87CEEB ALIGN="CENTER" colspan=4><font face="Arial">  American Fund </font></td></tr>
            <tr><td  bgcolor=#FFF8DC rowspan=3 colspan=2><font face="Arial"> Amount Invested:
            %(AMOUNT_INVESTED)s
            </font></td>
                <td  bgcolor=#FFFACD colspan=2><font face="Arial"> Initial Charge Ratio:
                %(INITIAL_CHARGE_RATIO)s
                </font></td></tr>
            <tr><td  bgcolor=#FFFACD colspan=2><font face="Arial"> Growth Rate:
            %(GROWTH_RATE)s
            </font></td></tr>
            <tr><td  bgcolor=#FFFACD colspan=2><font face="Arial"> Annual Charge Ratio:
            %(ANNUAL_CHARGE_RATIO)s
            </font></td></tr>
<tr><td  bgcolor=#FFF8DC align="right"><font face="Arial"> For year</font></td>
        <td  bgcolor=#FFFACD align=right><font face="Arial"> Invested to Date</font></td>
        <td  bgcolor=#FFFACD align=right><font face="Arial"> Total charges</font></td>
        <td  bgcolor=#FFFACD align=right><font face="Arial"> Valuation</font></td></tr>
        %(__content__)s
</table></CENTER>
"""
charges_row_format = """
<tr><td  bgcolor=#FFF8DC align="right"><font face="Arial">
                     %(YEAR)s
                     </font></td>
                        <td  bgcolor=#FFFACD align="right"><font face="Arial">
                        %(INVESTED_TO_DATE)s
                        </font></td>
                        <td  bgcolor=#FFFACD align="right"><font face="Arial">
                        %(EFFECT_OF_CHARGES)s
                        </font></td>
                        <td  bgcolor=#FFFACD align="right"><font face="Arial">
                        %(VALUATION)s</font></td></tr>
                        """

def currencystring(t):
    f = float(t)
    return "%10.2f" % f

def test():
    fn = "samples/hamlet.xml"
    ofn = "samples/hamlet.html"
    text = open(fn).read()
    M = MapController()
    # ref to %(TITLE.__content__)s should refer to the title of the play (not act or scene)
    M["PLAY"] = """ <html><head><title>%(TITLE.__content__)s</title></head><BODY>
                <CENTER><H1><FONT FACE="Arial">%(PLAY.TITLE.__content__)s</FONT></H1></CENTER>
         %(__content__)s
         <BODY>
         </html>
         """
    M[""] = "%(__content__)s"
    # TITLE VANISHES IT'S __CONTENT__ IS USED IN THE TOP LEVEL SUBSTITUTION
    M["TITLE"] = ""
##    # top level title
##    M["TITLE"] = """<HEAD>
##                <TITLE>%(__content__)s</TITLE>
##                </HEAD>
##                <BODY>
##                <CENTER><H1><FONT FACE="Arial">%(__content__)s</FONT></H1></CENTER>
##                """
    M["FM"] = """<BR><BR><BR>
                 <CENTER><TABLE><TR><TD BGCOLOR="#99A9FF">
                 %(__content__)s
                 </TD></TR></TABLE></CENTER>
                 """
    M["P"] = """<BR>%(__content__)s<BR>\n"""
    M["PERSONAE"] = """<BR><BR><BR>
                 <CENTER><TABLE>
                 %(__content__)s
                 </TABLE></CENTER>
                 """
    # override title
    M["PERSONAE"]["TITLE"] = "<CENTER><H3>%(PLAY.TITLE.__content__)s %(__content__)s</H3></CENTER>"
    M["PERSONA"] = """<TR><TD BGCOLOR="#A9FF99">%(__content__)s</TD></TR>\n"""
    M["PGROUP"] = """<TR><TD><HR></TR></TD>
                    %(__content__)s
                    <TR><TD><HR></TR></TD>"""
    M["GRPDESCR"] = """<TR><TD ALIGN="RIGHT"><EM>%(__content__)s</EM></TD></TR>"""
    M["SCNDESCR"] = """<CENTER><B>%(__content__)s</B></CENTER>\n"""
    M["PLAYSUBT"] = """<HR><BR><BR>%(__content__)s<BR><BR><HR>\n"""
    M["ACT"] = """%(__content__)s"""
    # override title behaviour
    M["ACT"]["TITLE"] = "<CENTER>%(PLAY.TITLE.__content__)s <H2>%(__content__)s</H2></CENTER>"
    M["SCENE"] = """%(__content__)s"""
    # override title behaviour
    M["SCENE"]["TITLE"] = "<CENTER><H3>%(PLAY.TITLE.__content__)s %(__content__)s</H3></CENTER>"
    THEAD = """<TABLE BORDER="0" WIDTH="98%%"><TR>\n"""
    TTAIL = """\n</TR></TABLE>\n"""
    M["STAGEDIR"] = THEAD + """<TD WIDTH="30%%"></TD><TD ALIGN=RIGHT><em>%(__content__)s</em></TD>\n""" + TTAIL
    M["SPEECH"] = THEAD + "%(__content__)s\n" + TTAIL + "<BR><BR>"
    # A stagedir inside a speech
    M["SPEECH"]["STAGEDIR"] = """<TD></TD><TD ALIGN=RIGHT><em>%(__content__)s</em></TD></TR><TR>\n"""
    M["SPEAKER"] = """<TD width="30%%"><B>%(__content__)s</B>\n</TD><TD></TD></TR><TR>"""
    M["LINE"] = """<TD></TD><TD>%(__content__)s</TD></TR><TR>"""
    M["LINE"]["STAGEDIR"] = "<EM>[%(__content__)s]</EM>"
    import os
    os.chdir('samples')
    otext = M.process(text)
    os.chdir('..')
    open(ofn, "w").write(otext)
    print("wrote", ofn)

if __name__=="__main__":
    from time import time
    now = time()
    print('processing samples/hamlet.xml')
    test()
    print("elapsed", time()-now)



# -*- coding: cp1252 -*-

"""
Portions copyright © 2005-2009, Stephen John Machin, Lingfo Pty Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. None of the names of Stephen John Machin, Lingfo Pty Ltd and any
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
/*-
 * Copyright (c) 2001 David Giffin.
 * All rights reserved.
 *
 * Based on the the Java version: Andrew Khan Copyright (c) 2000.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by
 *     David Giffin <david@giffin.org>."
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by
 *     David Giffin <david@giffin.org>."
 *
 * THIS SOFTWARE IS PROVIDED BY DAVID GIFFIN ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DAVID GIFFIN OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
"""
